/*
 GAMP - (c) 2001 - Renato R. Santana <rrsantana@yahoo.com>
 
 This is the effects module.
   
 This file is part of GAMP, the GNU Amiga Module Player
 
 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2, or (at your option)
 any later version.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
    
 You should have received a copy of the GNU General Public License
 along with This program; see the file COPYING.  If not, write to
 the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <stdio.h>
#include <unistd.h>

#include "effects.h"
#include "modctl.h"

#define 	NUM_OF_PERIODS		36

/* Original Amiga ProTracker periods */
static unsigned short       octaves[NUM_OF_PERIODS] = {
		856,808,762,720,678,640,604,570,538,508,480,453,
		428,404,381,360,339,320,302,285,269,254,240,226,
		214,202,190,180,170,160,151,143,135,127,120,113
/*
		850,802,757,715,674,637,601,567,535,505,477,450,
		425,401,379,357,337,318,300,284,268,253,239,225,
		213,201,189,179,169,159,150,142,134,126,119,113 */
};

unsigned char 		sintab[32]={
					0, 24, 49, 74, 97,120,141,161,
					180,197,212,224,235,244,250,253,
					255,253,250,244,235,224,212,197,
					180,161,141,120, 97, 74, 49, 24
					};

extern struct modctl 	modctl;

int get_period_idx(int period) {
	int		i;
	
	for(i = 0; i < NUM_OF_PERIODS; i++)
		if (octaves[i] == period)
			break;
	return(i);
}

int volume_range(int vol) {
	vol = vol > 64 ? 64 : vol;
	vol = vol <  0 ?  0 : vol;
	return(vol);
}

void effect_arpeggio(struct channeltype *ch, int ef_num, int ef_arg, int period) {
	int n;
	if (modctl.tick == 0) {
		ch->effect = ef_num;
		if (ef_arg && ch->current_period) {
			ch->effect_argx = (ef_arg & 0xf0) >> 4;
			ch->effect_argy = ef_arg & 0x0f;
			ch->effect_arp_ctr = 0;
			ch->effect_arp_orig_period_idx = get_period_idx(ch->current_period); 
		}
	}
	else {
		if (++ch->effect_arp_ctr > 2)
			ch->effect_arp_ctr = 0;
		if (ch->effect_arp_ctr == 0)
			ch->final_period = octaves[ch->effect_arp_orig_period_idx];
		else if (ch->effect_arp_ctr == 1)
			ch->final_period = octaves[ch->effect_arp_orig_period_idx + ch->effect_argx];
		else if (ch->effect_arp_ctr == 2)
			ch->final_period = octaves[ch->effect_arp_orig_period_idx + ch->effect_argy];
		reset_note(ch);
	}
}

void effect_porta_up(struct channeltype *ch, int ef_num, int ef_arg, int period) {
	int n;

	if (modctl.tick == 0) {
		ch->effect = ef_num;
		if (ef_arg)
			ch->effect_porta_up_speed = ef_arg;
	}
	else {
		ch->final_period -= ch->effect_porta_up_speed;
		ch->final_period = ch->final_period < octaves[NUM_OF_PERIODS] ? octaves[NUM_OF_PERIODS] : ch->final_period;
		reset_note(ch);
	}
}

void effect_porta_down(struct channeltype *ch, int ef_num, int ef_arg, int period) {
	int n;

	if (modctl.tick == 0) {
		ch->effect = ef_num;
		if (ef_arg)
			ch->effect_porta_down_speed = ef_arg;
	}
	else {
		ch->final_period += ch->effect_porta_down_speed;
		ch->final_period = ch->final_period > octaves[0] ? octaves[0] : ch->final_period;
		reset_note(ch);
	}
}

void effect_porta_to_note(struct channeltype *ch, int ef_num, int ef_arg, int period) {
	int n;

	if (modctl.tick == 0) {
		ch->effect = ef_num;
		if (period)
			ch->effect_note_to_porta_to = period;
		if (ef_arg)
			ch->effect_porta_speed = ef_arg;
	}
	else {
		n = ch->effect_note_to_porta_to > ch->final_period ? 1 : -1;
		ch->final_period += ch->effect_porta_speed * n;
		if (n > 0)
			ch->final_period = ch->final_period > ch->effect_note_to_porta_to ? ch->effect_note_to_porta_to : ch->final_period;
		else
			ch->final_period = ch->final_period < ch->effect_note_to_porta_to ? ch->effect_note_to_porta_to : ch->final_period;
		reset_note(ch);
	}
}

void effect_vibrato(struct channeltype *ch, int ef_num, int ef_arg, int period) {
	int				vib;

	if (modctl.tick == 0) {
		ch->effect = ef_num;
		ch->effect_vibrato_base_period = ch->final_period;
		if (ef_arg) {
			ch->effect_vibrato_speed = (ef_arg & 0xf0) >> 4;
			ch->effect_vibrato_depth = ef_arg & 0x0f; 
		}
		if (period) { 						/* if we have a new period, */
			ch->effect_vibrato_neg = 1;		/* we reset vibrato else we */
			ch->effect_vibrato_idx = 0;		/* continue from last position */
		}									/* on sintab */
	}
	else {
		vib = (ch->effect_vibrato_depth * sintab[ch->effect_vibrato_idx]) >> 7;
		ch->final_period = ch->effect_vibrato_base_period + vib * ch->effect_vibrato_neg;
//		ch->period += vib * ch->effect_vibrato_neg;
		ch->effect_vibrato_idx += ch->effect_vibrato_speed;
		if (ch->effect_vibrato_idx > 31) {
			ch->effect_vibrato_idx -= 32;
			ch->effect_vibrato_neg *= -1;    /* invert signal */
		}
		reset_note(ch);
	}
}

void effect_tremolo(struct channeltype *ch, int ef_num, int ef_arg, int period) {
	int				trem;

	if (modctl.tick == 0) {
		ch->effect = ef_num;
		ch->effect_tremolo_base_vol = ch->loudness;
		if (ef_arg) {
			ch->effect_tremolo_speed = (ef_arg & 0xf0) >> 4;
			ch->effect_tremolo_depth = ef_arg & 0x0f; 
		}
		if (period) { 						/* if we have a new period, */
			ch->effect_tremolo_neg = 1;		/* we reset tremolo, else we */
			ch->effect_tremolo_idx = 0;		/* continue from last position */
		}
	}
	else {
		trem = (ch->effect_tremolo_depth * sintab[ch->effect_tremolo_idx]) >> 6;
		ch->loudness = ch->effect_tremolo_base_vol + trem * ch->effect_tremolo_neg;
		ch->loudness = volume_range(ch->loudness);
		
		ch->effect_tremolo_idx += ch->effect_tremolo_speed;
		if (ch->effect_tremolo_idx > 31) {
			ch->effect_tremolo_idx -= 32;
			ch->effect_tremolo_neg *= -1;    /* invert neg value */
		}
	}
}

void effect_volume_slide(struct channeltype *ch, int ef_num, int ef_arg) {

	if(modctl.tick == 0) {
		ch->effect = ef_num;
		ch->effect_argx = (ef_arg & 0xf0) >> 4; /* slide up */
		ch->effect_argy = ef_arg & 0x0f;		/* slide down */
	}
	else {
		if (! ch->effect_argx * ch->effect_argy) { /* check if at least one is non-zero */
			ch->loudness += ch->effect_argx - ch->effect_argy;
			ch->loudness = volume_range(ch->loudness);
		}
	}
}

void process_effect(
struct channeltype *ch, int ef_num, int ef_arg, int period) {

	if (ef_num | ef_arg) {
		switch (ef_num) {
			case 0x00: {			/* arpeggio */
				if (ef_arg)
					effect_arpeggio(ch, ef_num, ef_arg, period);
				break;				   
			}
			case 0x01: {			/* porta up */
				effect_porta_up(ch, ef_num, ef_arg, period);
				break;
			}
			case 0x02: {			/* porta down */
				effect_porta_down(ch, ef_num, ef_arg, period);
				break;
			}
			case 0x03: {			/* porta to note */
				effect_porta_to_note(ch, ef_num, ef_arg, period);
				break;
			}
			case 0x04: {			/* vibrato */
				effect_vibrato(ch, ef_num, ef_arg, period);
				break;
			}
			case 0x05: {			/* porta to note + volume slide */
				effect_volume_slide(ch, ef_num, ef_arg);
				break;
			}
			case 0x06: {			/* vibrato + volume slide */
				effect_vibrato(ch, ef_num, 0, period);
				effect_volume_slide(ch, ef_num, ef_arg);
				break;
			}
			case 0x07: {			/* tremolo */
				effect_tremolo(ch, ef_num, ef_arg, period);
				break;
			}
			case 0x09: {			/* instrument offset */
				if (period && ef_arg << 8 < ch->instr_len)
					ch->playing_pos += ef_arg << 8;
				break;
			}
			case 0x0a: {			/* volume slide */
				effect_volume_slide(ch, ef_num, ef_arg);
				break;
			}
			case 0x0b: {			/* jump to pattern order */
				modctl.pos = ef_arg;
				modctl.pos = modctl.pos > modctl.maxpos ? modctl.maxpos : modctl.pos;
				modctl.row = 0;
				modctl.leaving_block = 1; /* no row increment */
				break;
			}
			case 0x0c: { 			/* set volume */
				ch->loudness = ef_arg;
				break;
			}
			case 0x0d: { 			/* pattern break */
				/* the parameter is a decimal number but stored as hexa */
				ef_arg = ((ef_arg & 0xf0) >> 4) * 10 + (ef_arg & 0x0f);
				ch->effect = ef_num;
				if(!modctl.leaving_block)
					modctl.pos++;
				modctl.row = ef_arg > 63 ? 0 : ef_arg;
				modctl.leaving_block = 1; /* causes no row increment */
				break;
			}
			case 0x0f: {  			/* change tempo (defaults are 6 & 125) */
				if (ef_arg) {
					if (ef_arg <= 31)			/* want to update speed */
						modctl.speed = ef_arg;
					else						/* want to update bpm */
						modctl.bpm = ef_arg;	
					reset_speed();
				}
				break;
			}
		
		}
	}
}


void effect_tick(struct channeltype *ch) {
	int 	n;
	
	switch (ch->effect) {
		case 0x00: {					/* arpeggio */
			if (ch->effect_argx + ch->effect_argy)
				effect_arpeggio(ch, 0, 0, 0);
			break;
		}
		case 0x01: {					/* porta up */
			effect_porta_up(ch, 0, 0, 0);
			break;
		}
		case 0x02: {					/* porta down */
			effect_porta_down(ch, 0, 0, 0); 
			break;
		}
		case 0x03: {					/* porta to note */
			effect_porta_to_note(ch, 0, 0, 0); 
			break;
		}
		case 0x04: {					/* vibrato */
			effect_vibrato(ch, 0, 0, 0);
			break;
		}
		case 0x05: {					/* porta to note + volume slide */
			effect_porta_to_note(ch, 0, 0, 0);
			effect_volume_slide(ch, 0, 0);
			break;
		}
		case 0x06: {					/* vibrato + volume slide */
			effect_vibrato(ch, 0, 0, 0);
			effect_volume_slide(ch, 0, 0);
			break;
		}
		case 0x07: {					/* tremolo */
			effect_tremolo(ch, 0, 0, 0);
			break;
		}
		case 0x0a: {					/* volume slide */
			effect_volume_slide(ch, 0, 0);
			break;
		}
	}
}

